/*
  mset.c - implementation of the mset routines
  see modkptrs.c for an explaination for msets and there use.

  initial implemenation assumes that log_2(K) <= 6 and thus I can use
  a single 64 bit longword as a bit vector to represent the set
  values.  This restriction may be relaxed if can be shown that larger
  k may increase the precision of the pointer analysis.

  #defines -

  MORE_PRECISE - I've never tried this.  Attempts to keep more
  precision by reverting to the basep representation when going to
  top.  Doesn't seem to matter so far.

  QUESTIONS:
  
  o can I check the result of mset operations before returning and if
  it is < Z_k, ?> return <{0}, I>?  (MORE_PRECISE)

  o Robert says it is "dangerous" to assume that registers are defined
  before they are used  (but it never happens to me...)

  NOTES:

  *** Warning *** don't forget that source and destination parameters
  may be aliased!!!

  The names for variable and operations were defined before the theory
  was properly formulated.  Therefore the names may seem archaic.  I
  choose to name the tuple <S, i> to be an "mset", Where S is a subset
  of Z_k (inters modulo k), and i is in of instruction pointers or {
  TOP, BOT }.  Thus an mset correspons to an "address descriptor", the
  basep field is the instruction component and the set field is the
  "modulo k set" component.  In the past, I choose as the exponent for
  our modulus rather than the modulus iteself.  (Thus the name
  mod2kptrs etc.)  Now K is the modulus.  The current definitions
  should match those specified by email message from debray to me
  (weippert) dated 7 May 1997 21:02, subject: monotonicity problem.

  i: ra   +  rb    ->  rc           MORE_PRECISE
  <Sa, B>    <Sb, B>   <Sc, B>      if Sc = Z_k then <{0}, i>
  <Sa, ia>   <Sb, B>   <Sc, ia>     if Sc = Z_k then <{0}, i>
  <Sa, ia>   <Sb, ib>  <{0}, i>     
  <Sa, T>    <Sb, *>   <{0}, i>

  Sc = { (a + b) mod k | all a in Sa, b in Sb }
  
  i: ra   *  rb   ->  rc
  <Sa, B>    <Sb, B>   <Sc, B>      if Sc = Z_k then <{0}, i>
  <Sa, ia>   <Sb, B>   <{0}, i>     <Sc', B> 
  <Sa, ia>   <Sb, ib>  <{0}, i>     
  <Sa, T>    <Sb, *>   <{0}, i>

  Sc = { (a*b) mod k | all a in Sa, b in Sb }
  Sc' = { (a*b) mod k | all a in Z_k, b in Sb }
  
  (below is from debray's email...)
  TOP == < Z_k, top >         ( "totally unknown" )
  BOT == < {},  bottom >      ( "nothing at all" )

  The union operation should now be defined as

  union(X, Y) =
     if (X == BOT) then Y
     else if (Y == BOT) then X
     else 
       let X = <m1, v1>, Y = <m2, v2>;
           if (v1 == v2) then < m1 u m2, v1 >
	   else TOP


  */
#include "global.h"
#include "functions.h"
#include "mod2kptrs.h"
#include "mset.h"

BOOL sameBase(INDEX *d, INDEX b1, INDEX b2);

BOOL insDominates(INDEX i1, INDEX i2) {
  INDEX i;
  
  if (AINS_BBL(i1) != AINS_BBL(i2) && Dominates(AINS_BBL(i1), AINS_BBL(i1)))
    /* from different basic blocks and b1 dominates b2 */
    return TRUE;
  
  else if (AINS_BBL(i1) == AINS_BBL(i2)) {
    /* from the same basic block ... */
    for(i = AINS_NEXT(i1); VALID(i); i = AINS_NEXT(i))
      if (i == i2)
	/* ... and i1 preceeds i2 */
	return TRUE;
    /* ... but i1 doesn't preceed i2 */
    return FALSE;
  }
  
  /* from different basic blocks b1 doesn't dominate b2 */
  return FALSE;
} /* insDominates */


/* determines 1. m1 and m2 have the same base pointer and 2. if the
   value in the registers defined by this instruction is seen the same
   from i1 and i2 */
BOOL mset_sameValue(mset *m1, INDEX i1, mset *m2, INDEX i2) {
  INDEX tmp;
  INDEX i0;
  
  if (m1->basep != m2->basep)
    /* the mod k values in each memset are relative to different
       unknown values */
    return FALSE;

  if ((i0 = m1->basep) == BASE_BOT)
    /* both are absolute (i.e. value == 0) */
    return TRUE;
    
  /* -- m1 and m2 are defined in terms of the same instruction, i0 -- */
  
  if (!insDominates(i1, i2)) {
    /* flip 'em */
    tmp = i1;
    i1 = i2;
    i2 = tmp;
  }
  /* -- i1 does not follow i2 -- */
  
  if (!insDominates(i0, i1) || !insDominates(i1, i2))
    /* m1 and m2 are defined in terms of a possibly different base value */
    return FALSE;

  /* -- base value for m1 and m2 are equal -- */
  return TRUE;

} /* mset_sameValue */


/* returns true if the operation changes d */
BOOL mset_copy(mset *d, mset *s) {
  BOOL change = d->set != s->set || d->basep != s->basep;

  if (change) {
    d->set = s->set;
    d->basep = s->basep;
  }
  return change;
} /* mset_copy */


/* sets the base value as that defined by instruciton bp */
void mset_base(mset *m, INDEX bp) {
    m->set = BITV_SINGLETON(0);
    m->basep = bp;
} /* mset_base */


/* sets m to MSET_TOP */
void mset_all(mset *m) {
  m->set = BITV_ALL;
  m->basep = BASE_TOP; 
} /* mset_all */


/* set ms to MSET_BOT */
void mset_empty(mset *d) {
  d->set = BITV_EMPTY;
  d->basep = BASE_BOT;
} /* mset_empty */

  
BOOL mset_union(mset *d, mset *m1, mset *m2) {
  mset r;

  if (MSET_IS_BOT(m1))
    return mset_copy(d, m2);

  if (MSET_IS_BOT(m2))
    return mset_copy(d, m1);

  if (m1->basep == m2->basep) {
    r.set = BITV_UNION(m1->set, m2->set);
    r.basep = m1->basep;
    return mset_copy(d, &r);
  }

  r.basep = BASE_TOP;
  r.set = BITV_ALL;
  return mset_copy(d, &r);
} /* mset_union */


#undef __PROC__
#define __PROC__ "mset_singleton"

void mset_singleton(mset *d, UINT64 v) {
  d->set = BITV_SINGLETON(MODK(v));
  d->basep = BASE_BOT;
}


/* assumes m1, m2 are defined relative to the same value (check with
   mset_saveValue) */
BOOL mset_disjoint(mset *m1, mset *m2) {
  return BITV_INTERSECT(m1->set, m2->set) == BITV_EMPTY;
}


BOOL mset_equal(mset *m1, mset *m2) {
  return m1->basep == m2->basep && m1->set == m2->set;
}


/* this operation essential implements a 64 bit logical left rotate */
void mset_addi(mset *d, mset *s, UINT64 v, INDEX i) {
  UINT64 highbits;
  UINT64 saveBitMask;

  if (s->basep == BASE_TOP) {
    mset_base(d, i);
    return;
  }
  
  v = MODK(v);

  /* save high bits that will get shifted off */
  saveBitMask = ((1 << v) - 1);
  highbits = s->set & (saveBitMask << (64-v));
  highbits >>= 64 - v;

  d->set = s->set << v;
  d->set |= highbits;
  d->basep = s->basep;

#ifdef MORE_PRECISE
  if (d->set == BITV_ALL)
    mset_base(d, i);
#endif
} /* mset_addi */


void mset_subi(mset *d, mset *s, UINT64 v, INDEX i) {
  mset_addi(d, s, MODULUS - MODK(v), i);
} /* mset_subi */
    

void mset_add(mset *d, mset *m1, mset *m2, INDEX i) {
  mset r;
  mset tmp;
  mset_value_t m;

  if (!sameBase(&r.basep, m1->basep, m2->basep)) {
    mset_base(d, i);
    return;
  }

  r.set = BITV_EMPTY;
  BITV_FOREACH(m2->set, m) {
    mset_addi(&tmp, m1, m, i);
    r.set = BITV_UNION(r.set, tmp.set);
  }

  if (r.set == BITV_EMPTY)
    r.basep = BASE_BOT;
  
#ifdef MORE_PRECISE
  if (r.set == BITV_ALL)
    mset_base(d, i);
  else
#endif
    mset_copy(d, &r);
} /* mset_add */


/* NOTE: same as add except use subi */
void mset_sub(mset *d, mset *m1, mset *m2, INDEX i) {
  mset r;
  mset tmp;
  mset_value_t m;

  if (!sameBase(&r.basep, m1->basep, m2->basep)) {
    mset_base(d, i);
    return;
  }

  r.set = BITV_EMPTY;
  BITV_FOREACH(m2->set, m) {
    mset_subi(&tmp, m1, m, i);
    r.set = BITV_UNION(r.set, tmp.set);
  }

  if (r.set == BITV_EMPTY)
    r.basep = BASE_BOT;
  
#ifdef MORE_PRECISE
  if (r.set == BITV_ALL)
    mset_base(d, i);
  else
#endif
    mset_copy(d, &r);
} /* mset_sub */


void mset_muli(mset *d, mset *s, UINT64 v, INDEX i) {
  mset_value_t m;
  mset_value_t t;
  mset news;  /* may have to use different source */

  if (s->basep != BASE_BOT) {
    /* multiply doesn't work with non zero base value
       modify source to be everything 
     */
#ifndef MORE_PRECISE
    mset_base(d, i);
    return;
#else
    news.set = BITV_ALL;
    news.basep = BASE_BOT;
#endif
  } /* if */
  else {
    /* source is absolute use input set s */
    news.set = s->set;
    news.basep = BASE_BOT;
  } /* else */
  v = MODK(v);
  
  d->set = BITV_EMPTY;
  d->basep = BASE_BOT;
  BITV_FOREACH(news.set, m) {
    t = MODK(m*v);
    d->set = BITV_ADD(d->set, t);
  } /* BITV_FOREACH */

  if (d->set == BITV_ALL)
    mset_base(d, i);
} /* mset_multi */


void mset_mul(mset *d, mset *m1, mset *m2, INDEX i) {
  mset r, tmp;
  mset a, b;
  mset_value_t m;

  mset_copy(&a, m1);
  mset_copy(&b, m2);
#ifndef MORE_PRECISE
  if (VALID_BASE(a.basep) || VALID_BASE(b.basep) ||
      a.basep == BASE_TOP || b.basep == BASE_TOP) {
    /* multiply doesn't work with non zero base value */
    mset_base(d, i);
    return;
  }
#else
  if ((VALID_BASE(a.basep) && VALID_BASE(b.basep)) || 
      a.basep == BASE_TOP || b.basep == BASE_TOP) {
    /* multiply doesn't work with non zero base value */
    mset_base(d, i);
    return;
  }

  /* -- both base pointers aren't valid instructions
     and neither are TOP -- */
  if (VALID_BASE(a.basep)) {
    a.basep = BASE_BOT;
    a.set = BITV_ALL;
  }

  if (VALID_BASE(b.basep)) {
    mset_copy(&b, &a);
    
    a.basep = BASE_BOT;
    a.set = BITV_ALL;
  }  
#endif /* #else */
  
  /* -- a is Z_k or neither a or b is Z_k -- */
  mset_empty(&r);
  BITV_FOREACH(b.set, m) {
    mset_muli(&tmp, &a, m, i);
    mset_union(&r, &tmp, &r);
  } /* BITV_FOREACH */    

  mset_copy(d, &r);

#ifdef MORE_PRECISE
  if (r.set == BITV_ALL)
    mset_base(&r, i);
#endif
} /* mset_mul */


/* checks to see if b1 and b2 are compatible base pointers
   returns the correct basepointer in d 
 */
BOOL sameBase(INDEX *d, INDEX b1, INDEX b2) {

  if (b1 == BASE_TOP || b2 == BASE_TOP)
    return FALSE;

  /* -- neither are TOP -- */
  if (b1 == BASE_BOT) {
    *d = b2;
    return TRUE;
  }

  /* -- b1 is a valid instr -- */
  if (b2 == BASE_BOT) {
    *d = b1;
    return TRUE;
  }

  /* -- b1, b2 are both valid instrs -- */
  return FALSE;
} /* sameBase */


void mset_printem(FILE *fp, mset *s) {
  mset_value_t m;

  if (fp == NULL)
    fp = stderr;
  
  if (VALID_BASE(s->basep))
    fprintf(fp, "(base %6d) ", s->basep);
  else
    fprintf(fp, "(base %6s) ", s->basep == BASE_TOP ? "TOP" : "BOT");
  if (s->set == BITV_ALL)
    fprintf(fp, "(everything)");
  else if (s->set == BITV_EMPTY)
    fprintf(fp, "(empty)");
  else
    BITV_FOREACH(s->set, m)
      fprintf(fp, "%2d ", m);
  putc('\n', fp);

} /* mset_printem */
