/*=======================================================
  Author:   R. Muth
  Version:  $Id: optimize.c,v 1.72 1998/12/10 00:14:11 muth Exp $
  Comment:  This file contains the optimization code
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

/*=======================================================
@OPTIMIZATION
Eliminate No-ops
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptEliminateNoops"

extern BOOL OptEliminateNoops(FILE *fp)
{
    INDEX  ibbl;
    INDEX  iins;
    BOOL   change;
    NUMBER kcount=0;
    NUMBER kcounth=0;
    INT64  weight=0;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Noop");
    clock_t time = clock();
    
    FOREACH_LIVE_INS(iins)
    {
        ibbl = AINS_BBL(iins);
        
        if( !AinsHasSideEffect(iins) &&
            (AINS_REG_C(iins) == REG_ZERO || AINS_REG_C(iins) == REG_FZERO)) 
        {
            if( verbose )
            {
                fprintf(fp,"\tKilling: ");
                PrintIns(fp,iins);
            }
            kcount++;
            if( ABBL_IS_HOT(ibbl) )
                kcounth++;
            weight += ABBL_COUNT(ibbl);
            
            AinsKill(iins);
        }
    }


    FOREACH_LIVE_INS(iins)
    {
        ibbl = AINS_BBL(iins);
        
        if( AinsIsNoop(iins) )
        {
            if( verbose )
            {
                fprintf(fp,"\tKilling: ");
                PrintIns(fp,iins);
            }
            kcount++;
            if( ABBL_IS_HOT(ibbl) )
                kcounth++;
            weight += ABBL_COUNT(ibbl);
            
            AinsKill(iins);
        }
    }
        
    /* kill useless moves */

    FOREACH_LIVE_INS(iins)
    {
        ibbl = AINS_BBL(iins);
                
        if( AINS_CODE(iins) == I_lda &&
            AINS_DATA(iins) == 0 &&
            AINS_REG_B(iins) == AINS_REG_C(iins) )
        {
            if( verbose )
            {
                fprintf(fp,"\tKilling: ");
                PrintIns(fp,iins);
            }
            kcount++;
            if( ABBL_IS_HOT(AINS_BBL(iins)) )
                kcounth++;
            weight += ABBL_COUNT(ibbl);
            
            AinsKill(iins);
        }

        if( AINS_CODE(iins) == I_cpys &&
            AINS_REG_B(iins) == AINS_REG_C(iins) &&
            AINS_REG_B(iins) == AINS_REG_A(iins) )
        {
            if( verbose )
            {
                fprintf(fp,"\tKilling: ");
                PrintIns(fp,iins);
            }
            kcount++;
            if( ABBL_IS_HOT(AINS_BBL(iins)) )
                kcounth++;
            weight += ABBL_COUNT(ibbl);
            
            AinsKill(iins);
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (kcount>0);

    InfoStatus(fp,"\t[OptNoop%c] Killed Noops %d#%d  wt: %ld {%d ms}\n",CHANGE(change),kcount,kcounth,weight,time);

    return change;
}

/*=======================================================
@OPTIMIZATION
Eliminate Unreachable Code .
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptRemoveUnreachableCode"

#define ABBL_MAGIC(i) ABBL_U1(i).doubleword.d1

extern BOOL OptRemoveUnreachableCode(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Reach");
    BOOL   change = 0;
    INDEX iedg;
    INDEX ibbl;
    INDEX ifun;
    BOOL  flag;
    NUMBER rcount;
    NUMBER icount;
    NUMBER bcount;
    NUMBER fcount;
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) )
            ABBL_MAGIC(ibbl) = 1;
        else
            ABBL_MAGIC(ibbl) = 0;
    }
    
    rcount = 0;
    do
    {
        flag = 0;
        rcount++;
        
        FOREACH_LIVE_BBL(ibbl)
        {
            if( ABBL_MAGIC(ibbl) == 0 ) continue;

            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                INDEX jbbl = AEDG_TAIL(iedg);

                /* do not follow return edges,
                   the link edge will guarantee the survival */
                
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
                    continue;

                if( ABBL_MAGIC(jbbl) == 0 )
                {
                    ABBL_MAGIC(jbbl) = 1;
                    flag = 1;
                }
            }
        }
    }
    while(flag);

    icount = 0;
    bcount = 0;
    fcount = 0;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_MAGIC(ibbl) == 0 )
        {
            /* kill exit bbl  only if first */
            if( ABBL_TYPE(ibbl) == BT_EXIT &&  !ABBL_IS_FIRST(ibbl) )
                    continue;

            bcount++;
            icount += AbblNInss(ibbl);
            change = 1;            
#if 0
            CWARNING( CONFIG_USE_PROFILE && ABBL_COUNT(ibbl) > 0,("unreachable bbl %d has bbl count > 0\n",ibbl));
#endif            
            AbblKill(ibbl);
            if( verbose )
            {
                fprintf(fp,"\tKilling Bbl: %d\n",ibbl);
            }
        }
    }

    FOREACH_LIVE_FUN(ifun)
    {
        ibbl = AFUN_BBL_FIRST(ifun);
        if( VALID(ibbl) ) continue;

        fcount++;
        change = 1;            
        AfunKill(ifun);
        if( verbose )
        {
            fprintf(fp,"\tKilling Fun: %d [%s]\n",ifun,AFUN_NAME(ifun));
        }
        

    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[OptReach%c] Unreachable Funs: %d  Bbls: %d  Inss: %d (Rounds: %d) {%d ms}\n",
           CHANGE(change),fcount,bcount,icount,rcount,time);

    return change;
}

/*=======================================================
@OPTIMIZATION
Eliminate Empty Basic Blocks. 
  Remove empty basic blocks which must be a bt_norm
  basic block by forwarding its incoming edges to the next bbl
  Except if the basic block is a dummy entry basic block
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptUnlinkEmptyBbls"

extern BOOL OptUnlinkEmptyBbls(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Emptybbl");
    BOOL  change;
    INDEX ibbl;
    NUMBER ecount = 0;
    NUMBER ecounth = 0;
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        INDEX iedg;
        INDEX jedg;
        INDEX next;

        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
        if( ABBL_HAS_INSS(ibbl) ) continue;
        if( ABBL_TYPE(ibbl) == BT_EXIT ) continue;
        if( VALID(AFUN_MASTER(ABBL_FUN(ibbl))) ) continue;
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            if( AEDG_TYPE(iedg)==ET_FUN_LINK )
                break;            
        }
        if( VALID(iedg) ) continue;

        if( verbose )
        {
            fprintf(fp,"\t[*]  bbl: %d (%c)\n",ibbl,ABBL_TYPE(ibbl));
        }
        
        CASSERT( ABBL_TYPE(ibbl) == BT_NORM,("empty bbl is not bt_norm\n"));
        next = AEDG_TAIL(ABBL_EDG_NORM(ibbl));
        CASSERT( ABBL_TYPE(next) != BT_EXIT,("next bbl is exit bbl\n"));
        CERROR( AFUN_IS_PSEUDO(ABBL_FUN(next)),("next bbl is in PSEUDO\n"));

        AbblMigrateRelocInfo(ibbl,next);   /* very important */
        
        ecount++;
        if( ABBL_IS_HOT(ibbl) )
            ecounth++;

        /*
          the next statement avoids some warning messages
          if ibbl is the first bbl of a function.
          */

#if 0
        AbblUnlink(ibbl);
#endif
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            jedg = AedgNewFancy(AEDG_HEAD(iedg),next,AEDG_TYPE(iedg));
            AedgKillFancy(iedg);
            AedgLinkFancy(jedg);
        }

        AbblKill(ibbl);
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (ecount>0);

    InfoStatus(fp,"\t[OptEmptybbl%c] Unlinked Empty bbls: %d#%d {%d ms}\n",CHANGE(change),ecount,ecounth,time);
    
    return change;    
}
